from sklearn.utils.validation import check_array
import numpy as np
import numbers


def check_cu_co(cu, co, n_outputs):
    """Validate under- and overage costs.

    Parameters
    ----------
    cu : {ndarray, Number or None}, shape (n_outputs,)
       The underage costs per unit. Passing cu=None will output an array of ones.
    co : {ndarray, Number or None}, shape (n_outputs,)
       The overage costs per unit. Passing co=None will output an array of ones.
    n_outputs : int
       The number of outputs.
    Returns
    -------
    cu : ndarray, shape (n_outputs,)
       Validated underage costs. It is guaranteed to be "C" contiguous.
    co : ndarray, shape (n_outputs,)
       Validated overage costs. It is guaranteed to be "C" contiguous.
    """
    costs = [[cu, "cu"], [co, "co"]]
    costs_validated = []
    for c in costs:
        if c[0] is None:
            cost = np.ones(n_outputs, dtype=np.float64)
        elif isinstance(c[0], numbers.Number):
            cost = np.full(n_outputs, c[0], dtype=np.float64)
        else:
            cost = check_array(
                c[0], accept_sparse=False, ensure_2d=False, dtype=np.float64,
                order="C"
            )
            if cost.ndim != 1:
                raise ValueError(c[1], "must be 1D array or scalar")

            if cost.shape != (n_outputs,):
                raise ValueError("{}.shape == {}, expected {}!"
                                 .format(c[1], cost.shape, (n_outputs,)))
        costs_validated.append(cost)
    cu = costs_validated[0]
    co = costs_validated[1]
    return cu, co


def formate_hyperparameter(value, name, n_outputs):
    # make sure value is an array of shape (n_outputs,)
    if not isinstance(value, (list, tuple, np.ndarray)):
        value = np.full((n_outputs,), value)
    value = np.array(value)

    if value.ndim != 1:
        raise ValueError("%s must be 1D array or scalar" % (name))
    return value



from ._base import BaseNewsvendor, DataDrivenMixin
from ..utils.validation import check_cu_co
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.backend import switch, less
from tensorflow.keras.backend import sum as ksum
from tensorflow import cast

from sklearn.utils.validation import check_is_fitted
import numpy as np

ACTIVATIONS = ['elu', 'selu', 'linear', 'tanh', 'relu', 'softmax', 'softsign', 'softplus',
               'sigmoid', 'hard_sigmoid', 'exponential']


class DeepLearningNewsvendor(BaseNewsvendor, DataDrivenMixin):
    """A Deep-Learning model to solve the Newsvendor problem.

    This class implements the approach described in [1].

    Parameters
    ----------
    cu : {array-like of shape (n_outputs,), Number or None}, default=None
       The underage costs per unit. If None, then underage costs are one
       for each target variable
    co : {array-like of shape (n_outputs,), Number or None}, default=None
       The overage costs per unit. If None, then overage costs are one
       for each target variable
    neurons : list, default=[100,50]
        The ith element represents the number of neurons in the ith hidden layer
        Only used when hidden_layers='custom'.
    activations : list, default=['relu','relu']
        The ith element of the list represents the activation function of the ith layer.
        Valid activation functions are: 'elu', 'selu', 'linear', 'tanh', 'relu', 'softmax',
        'softsign', 'softplus','sigmoid', 'hard_sigmoid', 'exponential'.
        Only used when hidden_layers='custom'.
    optimizer: {'adam', 'sgd'}, default='adam'
        The optimizer to be used.
    epochs: int, default=100
        Number of epochs to train the model
    verbose: int 0, 1, or 2, default=0
        Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.

    Attributes
    ----------
    model_ : tensorflow.keras.Sequential
        The underlying model
    n_features_ : int
        The number of features when ``fit`` is performed.
    n_outputs_ : int
        The number of outputs.
    cu_ : ndarray, shape (n_outputs,)
        Validated underage costs.
    co_ : ndarray, shape (n_outputs,)
        Validated overage costs.

    References
    ----------
    .. [1] Afshin Oroojlooyjadid, Lawrence V. Snyder, Martin Takáˇc,
            "Applying Deep Learning to the Newsvendor Problem", 2018.

    Examples
    --------
    >>> from ddop.datasets import load_yaz
    >>> from ddop.newsvendor import DeepLearningNewsvendor
    >>> from sklearn.model_selection import train_test_split
    >>> X, Y = load_yaz(include_prod=['STEAK'],return_X_y=True)
    >>> cu,co = 15,10
    >>> X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=0.25, shuffle=False, random_state=0)
    >>> mdl = DeepLearningNewsvendor(cu, co)
    >>> mdl.fit(X_train, Y_train)
    >>> mdl.score(X_test, Y_test)
    TODO: ADD SCORE
    """

    def __init__(self, cu=None, co=None, neurons=[100, 50], activations=['relu', 'relu'], optimizer='adam', epochs=100,
                 verbose=0):
        self.neurons = neurons
        self.activations = activations
        self.optimizer = optimizer
        self.epochs = epochs
        self.verbose = verbose
        super().__init__(
            cu=cu,
            co=co)

    def _nv_loss(self, cu, co):
        """Create a newsvendor loss function with the given under- and overage costs"""

        def customized_loss(y_true, y_pred):
            y_true = cast(y_true, y_pred.dtype)
            loss = switch(less(y_pred, y_true), cu * (y_true - y_pred), co * (y_pred - y_true))
            return ksum(loss)

        return customized_loss

    def _create_model(self):
        """Create model"""
        neurons = self.neurons
        activations = self.activations
        n_features = self.n_features_
        n_outputs = self.n_outputs_

        model = Sequential()

        for size, activation in zip(neurons, activations):
            model.add(Dense(units=size, activation=activation))
        model.add(Dense(n_outputs))
        model.build((None, n_features))

        model.compile(loss=self._nv_loss(self.cu_, self.co_), optimizer=self.optimizer)

        return model

    def fit(self, X, y):
        """Fit the model to the training set (X, y).

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples, n_outputs)
            The target values.

        Returns
        ----------
        self : DeepLearningNewsvendor
            Fitted estimator
        """

        # Validate input parameters
        self._validate_hyperparameters()

        X, y = self._validate_data(X, y, multi_output=True)

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        # Determine output settings
        self.X_ = X
        self.y_ = y
        self.n_features_ = X.shape[1]
        self.n_outputs_ = y.shape[1]

        # Check and format under- and overage costs
        self.cu_, self.co_ = check_cu_co(self.cu, self.co, self.n_outputs_)
        model = self._create_model()
        model.fit(X, y, epochs=self.epochs, verbose=self.verbose)
        self.model_ = model
        return self

    def _validate_hyperparameters(self):
        """validate hyperparameters"""

        # Make sure self.neurons is a list
        neurons = self.neurons
        if not hasattr(neurons, "__iter__"):
            neurons = [neurons]
        neurons = list(neurons)

        # Make sure self.activations is a list
        activations = self.activations
        if not hasattr(activations, "__iter__"):
            activations = [activations]
        activations = list(activations)

        if np.any(np.array(neurons) <= 0):
            raise ValueError("neurons must be > 0, got %s." %
                             self.neurons)

        if np.any(np.array([activation not in ACTIVATIONS for activation in activations])):
            raise ValueError("Invalid activation function in activations. Supported are %s but got %s"
                             % (list(ACTIVATIONS), activations))

        if len(neurons) != len(activations):
            raise ValueError("Neurons and activations must have same length but neurons is of length %s and "
                             "activations %s " % (len(neurons), len(activations)))

        if self.verbose not in [0, 1, 2]:
            raise ValueError("verbose must be either 0, 1 or 2, got %s." %
                             self.verbose)

    def predict(self, X):
        """Predict values for X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        ----------
        y : array-like of shape (n_samples, n_outputs)
            The predicted values
        """
        check_is_fitted(self)
        pred = self.model_.predict(X)
        return pred



import unittest
from ddop.metrics._costs import pairwise_costs, average_costs, total_costs
from numpy.testing import assert_array_equal


class TestCosts(unittest.TestCase):

    def test_pairwise_costs(self):
        assert_array_equal(pairwise_costs([1], [10], 1, 2), [[18]])
        assert_array_equal(pairwise_costs([10], [1], 2, 1), [[18]])
        assert_array_equal(pairwise_costs([1, 10], [2, 5], 1, 1), [[1], [5]])
        assert_array_equal(pairwise_costs([1, 10], [2, 5], 2, 1), [[1], [10]])
        assert_array_equal(pairwise_costs([1, 10], [2, 5], 1, 2), [[2], [5]])
        # multioutput
        assert_array_equal(pairwise_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], 1, 1), [[1, 6], [1, 3]])
        # multioutput and differend cost coefficients
        assert_array_equal(pairwise_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], [1, 1], [1, 2]), [[1, 6], [1, 6]])

    def test_total_costs(self):
        assert_array_equal(total_costs([1], [10], 1, 2), 18)
        assert_array_equal(total_costs([10], [1], 2, 1), 18)
        assert_array_equal(total_costs([1, 10], [2, 5], 1, 1), 6)
        assert_array_equal(total_costs([1, 10], [2, 5], 2, 1), 11)
        assert_array_equal(total_costs([1, 10], [2, 5], 1, 2), 7)
        # multioutput
        assert_array_equal(total_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], 1, 1, multioutput="cumulated"), 11)
        assert_array_equal(total_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], 1, 1, multioutput="raw_values"), [2, 9])
        # multioutput and differend cost coefficients
        assert_array_equal(total_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], [1, 1], [1, 2], multioutput="cumulated"),
                           14)
        assert_array_equal(total_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], [1, 1], [1, 2], multioutput="raw_values"),
                           [2, 12])

    def test_average_costs(self):
        assert_array_equal(average_costs([1], [10], 1, 2), 18)
        assert_array_equal(average_costs([10], [1], 2, 1), 18)
        assert_array_equal(average_costs([1, 10], [2, 5], 1, 1), 3)
        assert_array_equal(average_costs([1, 10], [2, 5], 2, 1), 5.5)
        assert_array_equal(average_costs([1, 10], [2, 5], 1, 2), 3.5)
        # multioutput
        assert_array_equal(average_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], 1, 1, multioutput="uniform_average"), 2.75)
        assert_array_equal(average_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], 1, 1, multioutput="raw_values"), [1, 4.5])
        # multioutput and differend cost coefficients
        assert_array_equal(average_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], [1, 1], [1, 2], multioutput="uniform_average"),
                           3.5)
        assert_array_equal(average_costs([[1, 10], [2, 5]], [[2, 4], [1, 8]], [1, 1], [1, 2], multioutput="raw_values"),
                           [1, 6])


if __name__ == '__main__':
    unittest.main()



from ._costs import pairwise_costs, average_costs, total_costs, prescriptiveness_score
from ._scorer import make_scorer

__all__ = ["pairwise_costs", "total_costs", "average_costs", "prescriptiveness_score",
           "make_scorer"]



from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from abc import ABC, abstractmethod
from ..metrics._costs import average_costs
import numpy as np


class BaseNewsvendor(BaseEstimator, ABC):
    """
    Base class for newsvendor.
    """
    @abstractmethod
    def __init__(self, cu, co):
        self.cu = cu
        self.co = co


class DataDrivenMixin:
    _estimator_type = "regressor"
    def score(self, X, y):
        """
        Return the average costs of the prediction

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.
        y : array-like of shape (n_samples, n_outputs)
            The true values for x.

        Returns
        -------
         score: float
            The average costs
        """

        y_pred = self.predict(X)
        score = -average_costs(y, y_pred, self.cu_, self.co_, multioutput="uniform_average")
        return score


class ClassicMixin:
    _estimator_type = "regressor"
    def score(self, y, X=None):
        """
        Return the average costs of the prediction

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Exogenous variables are ignored
        y : array-like of shape (n_samples, n_outputs)
            The true values.

        Returns
        -------
         score: float
            The average costs
        """

        y = check_array(y, ensure_2d=False)

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        y_pred = self.predict(y.shape[0])
        return -average_costs(y, y_pred, self.cu_, self.co_, multioutput="uniform_average")



from ._SampleAverageApproximationNewsvendor import SampleAverageApproximationNewsvendor
from ._WeightedNewsvendor import DecisionTreeWeightedNewsvendor, RandomForestWeightedNewsvendor, \
    KNeighborsWeightedNewsvendor, GaussianWeightedNewsvendor
from ._LinearRegressionNewsvendor import LinearRegressionNewsvendor
from ._DeepLearningNewsvendor import DeepLearningNewsvendor

__all__ = ["SampleAverageApproximationNewsvendor", "DecisionTreeWeightedNewsvendor",
           "RandomForestWeightedNewsvendor", "KNeighborsWeightedNewsvendor",
           "GaussianWeightedNewsvendor", "LinearRegressionNewsvendor",
           "DeepLearningNewsvendor"]



from sklearn.metrics._scorer import _BaseScorer
from ..newsvendor import _SampleAverageApproximationNewsvendor as SAA
import inspect
import numpy as np


class _Scorer(_BaseScorer):
    def _score(self, method_caller, estimator, X, y_true):
        """Evaluate predicted target values for X relative to y_true.
        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.
        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.
        X : {array-like, sparse matrix}
            Test data that will be fed to estimator.predict.
        y_true : array-like
            Gold standard target values for X.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """

        y_pred = method_caller(estimator, "predict", X)
        cu = estimator.cu_
        co = estimator.co_

        if "y_pred_saa" in inspect.getfullargspec(self._score_func).args:
            X = estimator.X_
            y = estimator.y_
            y_pred_saa = SAA.SampleAverageApproximationNewsvendor(cu, co).fit(y_true).predict().flatten()
            y_pred_saa = np.full(y_true.shape, y_pred_saa)
            return self._sign * self._score_func(y_true, y_pred, y_pred_saa, cu, co, **self._kwargs)

        else:
            print("else")
            return self._sign * self._score_func(y_true, y_pred, cu, co, **self._kwargs)


def make_scorer(score_func, greater_is_better=True, **kwargs):
    """Make a scorer from a performance metric or loss function.
    This factory function wraps scoring functions for use in
    `sklearn.model_selection.GridSearchCV` and
    `sklearn.model_selection.cross_val_score`.
    It takes a score function from `ddop.metrics`, such as `ddop.metrics.total_costs`,
    and returns a callable that scores an estimator's output.
    The signature of the call is `(estimator, X, y)` where `estimator`
    is the model to be evaluated, `X` is the data and `y` is the
    ground truth labeling.

    Parameters
    ----------
    score_func : callable
        Score function included in ddop.metrics.
    greater_is_better : bool, default=True
        Whether score_func is a score function (default), meaning high is good,
        or a loss function, meaning low is good. In the latter case, the
        scorer object will sign-flip the outcome of the score_func.
    **kwargs : additional arguments
        Additional parameters to be passed to score_func.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score; greater is better.
    """
    sign = 1 if greater_is_better else -1

    cls = _Scorer

    return cls(score_func, sign, kwargs)


import pandas as pd
import numpy as np
from os.path import dirname, join
from sklearn.utils import Bunch


def load_yaz(include_date=False, one_hot_encoding=False,
             label_encoding=False, return_X_y=False):
    """Load and return the YAZ dataset

    Yaz is a fast casual restaurant in Stuttgart providing good service and food at short waiting times.
    The dataset contains the demand for the main ingredients at YAZ. Moreover, it stores a
    number of demand features. A description of targets and features is given below.

    **Dataset Characteristics:**

        :Number of Instances: 765

        :Number of Targets: 7

        :Number of Features: 12

        :Target Information:
            - 'calamari' the demand for calamari
            - 'fish' the demand for fish
            - 'shrimp' the demand for shrimps
            - 'chicken' the demand for chicken
            - 'koefte' the demand for koefte
            - 'lamb' the demand for lamb
            - 'steak' the demand for steak

        :Feature Information:
            - 'date' the date,
            - 'weekday' the day of the week,
            - 'month' the month of the year,
            - 'year' the year,
            - 'is_holiday' whether or not it is a national holiday,
            - 'is_closed' whether or not the restaurant is closed,
            - 'weekend' whether or not it is weekend,
            - 'wind' the wind force,
            - 'clouds' the cloudiness degree,
            - 'rain' the amount of rain,
            - 'sunshine' the sunshine hours,
            - 'temperature' the outdoor temperature,

    Parameters
    ----------
    include_date : bool, default=False
        Whether to include the demand date
    one_hot_encoding : bool, default=False
        Whether to one hot encode categorical features
    label_encoding : bool, default=False
        Whether to convert categorical columns (weekday, month, year) to continuous.
        Will only be applied if `one_hot_encoding=False`
    return_X_y : bool, default=False.
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.

    Returns
    -------
    data : sklearn Bunch
        Dictionary-like object, with the following attributes.

        data : Pandas DataFrame of shape (765, n_features)
            The data matrix.
        target: Pandas DataFrame of shape (765, n_targets)
            The target values.
        n_features: int
            The number of features included
        n_targets: int
            The number of target variables included
        DESCR: str
            The full description of the dataset.
        data_filename: str
            The path to the location of the data.
        target_filename: str
            The path to the location of the target.

    (data, target) : tuple if ``return_X_y`` is True

    Examples
    ---------
    >>> from ddop.datasets import load_yaz
    >>> X, y = load_yaz(return_X_y=True)
    >>> print(X.shape)
        (765, 11)
    """

    module_path = dirname(__file__)
    base_dir = join(module_path, 'data')
    data_filename = join(base_dir, 'yaz_data.csv')
    data = pd.read_csv(data_filename)
    target_filename = join(base_dir, 'yaz_target.csv')
    target = pd.read_csv(target_filename)

    with open(join(module_path, 'descr', 'yaz.rst')) as rst_file:
        fdescr = rst_file.read()

    if not include_date:
        data = data.drop('date', axis=1)

    n_features = data.shape[0]
    n_targets = data.shape[1]

    if one_hot_encoding:
        data = pd.get_dummies(data, columns=["weekday", "month", "year"])

    elif not one_hot_encoding and label_encoding:
        data['weekday'] = data['weekday'].apply(_day_to_continuouse)
        data['month'] = data['month'].apply(_month_to_continuouse)

    else:
        data['year'] = data['year'].apply(str)

    if return_X_y:
        return data, target

    return Bunch(data=data,
                 target=target,
                 n_features=n_features,
                 n_targets=n_targets,
                 DESCR=fdescr,
                 data_filename=data_filename,
                 target_filename=target_filename)


def load_bakery(include_date=False, one_hot_encoding=False,
             label_encoding=False, return_X_y=False):
    """Load and return the bakery dataset

    The bakery dataset contains the demand for a number of products from different stores. Moreover, it stores a
    number of demand features. A description of targets and features is given below.

    **Dataset Characteristics:**

        :Number of Instances: 127575

        :Number of Targets: 1

        :Number of Features: 13

        :Target Information:
            - 'demand' the corresponding demand observation

        :Feature Information:
            - 'date' the date
            - 'weekday' the day of the week,
            - 'month' the month of the year,
            - 'year' the year,
            - 'is_holiday' whether or not it is a national holiday,
            - 'is_holiday_next2days' whether or not it is a national holiday in the next two days,
            - 'is_schoolholiday' whether or not it is a school holiday,
            - 'store' the store id,
            - 'product' the product id,
            - 'rain' the amount of rain,
            - 'temperature' the average temperature in °C,
            - 'promotion_currentweek' whether or not there is a promotion this week
            - 'promotion_lastweek' whether there was a promotion last week

    Parameters
    ----------
    include_date : bool, default=False
        Whether to include the demand date
    one_hot_encoding : bool, default=False
        Whether to one hot encode categorical features
    label_encoding : bool, default=False
        Whether to convert categorical columns (weekday, month, year) to continuous.
        Will only be applied if `one_hot_encoding=False`
    return_X_y : bool, default=False.
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.

    Returns
    -------
    data : sklearn Bunch
        Dictionary-like object, with the following attributes.

        data : Pandas DataFrame of shape (127575, n_features)
            The data matrix.
        target: Pandas DataFrame of shape (127575, n_targets)
            The target values.
        n_features: int
            The number of features included
        n_targets: int
            The number of target variables included
        DESCR: str
            The full description of the dataset.
        data_filename: str
            The path to the location of the data.
        target_filename: str
            The path to the location of the target.
    (data, target) : tuple if ``return_X_y`` is True

    Examples
    ----------
    >>> from ddop.datasets import load_bakery
    >>> X, y = load_bakery(return_X_y=True)
    >>> print(X.shape)
        (127575, 12)
    """

    module_path = dirname(__file__)
    base_dir = join(module_path, 'data')
    data_filename = join(base_dir, 'bakery_data.csv')
    data = pd.read_csv(data_filename)
    target_filename = join(base_dir, 'bakery_target.csv')
    target = pd.read_csv(target_filename)

    with open(join(module_path, 'descr', 'bakery.rst')) as rst_file:
        fdescr = rst_file.read()

    if not include_date:
        data = data.drop('date', axis=1)

    n_features = data.shape[0]
    n_targets = data.shape[1]

    if one_hot_encoding:
        data = pd.get_dummies(data, columns=["weekday", "month", "year"])

    elif not one_hot_encoding and label_encoding:
        data['weekday'] = data['weekday'].apply(_day_to_continuouse)
        data['month'] = data['month'].apply(_month_to_continuouse)

    else:
        data['year'] = data['year'].apply(str)

    if return_X_y:
        return data, target

    return Bunch(data=data,
                 target=target,
                 n_features=n_features,
                 n_targets=n_targets,
                 DESCR=fdescr,
                 data_filename=data_filename,
                 target_filename=target_filename)


def load_SID(include_date=False, one_hot_encoding=False,
             label_encoding=False, return_X_y=False):

    """Load and return the store item demand dataset.

    **Dataset Characteristics:**

        :Number of Instances: 887284

        :Number of Targets: 1

        :Number of Features: 6

        :Target Information:
            - 'demand' the corresponding demand observation

        :Feature Information:
            - 'date' the date
            - 'weekday' the day of the week,
            - 'month' the month of the year,
            - 'year' the year,
            - 'store' the store id,
            - 'item' the item id
    Parameters
    ----------
    include_date : bool, default=False
        Whether to include the demand date
    one_hot_encoding : bool, default=False
        Whether to one hot encode categorical features
    label_encoding : bool, default=False
        Whether to convert categorical columns (weekday, month, year) to continuous.
        Will only be applied if `one_hot_encoding=False`
    return_X_y : bool, default=False.
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.

    Returns
    -------
    data : sklearn Bunch
        Dictionary-like object, with the following attributes.

        data : Pandas DataFrame of shape (887284, n_features)
            The data matrix.
        target: Pandas DataFrame of shape (887284, n_targets)
            The target values.
        n_features: int
            The number of features included
        n_targets: int
            The number of target variables included
        DESCR: str
            The full description of the dataset.
        data_filename: str
            The path to the location of the data.
        target_filename: str
            The path to the location of the target.
    (data, target): tuple if ``return_X_y`` is True

    Notes
    -----
    The store item demand dataset was published within a demand forecasting challenge on kaggle [1]

    References
    ----------
    .. [1] https://www.kaggle.com/c/demand-forecasting-kernels-only/overview

    Examples
    --------
    >>> from ddop.datasets import load_SID
    >>> X, y = load_SID(return_X_y=True)
    >>> print(X.shape)
        (887284, 5)
    """

    module_path = dirname(__file__)
    base_dir = join(module_path, 'data')
    data_filename = join(base_dir, 'SID_data.csv')
    data = pd.read_csv(data_filename)
    target_filename = join(base_dir, 'SID_target.csv')
    target = pd.read_csv(target_filename)

    with open(join(module_path, 'descr', 'SID.rst')) as rst_file:
        fdescr = rst_file.read()

    if not include_date:
        data = data.drop('date', axis=1)

    n_features = data.shape[0]
    n_targets = data.shape[1]

    if one_hot_encoding:
        data = pd.get_dummies(data, columns=["weekday", "month", "year"])

    elif not one_hot_encoding and label_encoding:
        data['weekday'] = data['weekday'].apply(_day_to_continuouse)
        data['month'] = data['month'].apply(_month_to_continuouse)

    else:
        data['year'] = data['year'].apply(str)

    if return_X_y:
        return data, target

    return Bunch(data=data,
                 target=target,
                 n_features=n_features,
                 n_targets=n_targets,
                 DESCR=fdescr,
                 data_filename=data_filename,
                 target_filename=target_filename)


def _month_to_continuouse(x):
    if x=='JAN':
        return 1
    elif x=='FEB':
        return 2
    elif x=='MAR':
        return 3
    elif x=='APR':
        return 4
    elif x=='MAY':
        return 5
    elif x=='JUN':
        return 6
    elif x=='JUL':
        return 7
    elif x=='AUG':
        return 8
    elif x=='SEP':
        return 9
    elif x=='OCT':
        return 10
    elif x=='NOV':
        return 11
    else:
        return 12


def _day_to_continuouse(x):
    if x=='MON':
        return 1
    elif x=='TUE':
        return 2
    elif x=='WED':
        return 3
    elif x=='THU':
        return 4
    elif x=='FRI':
        return 5
    elif x=='SAT':
        return 6
    else:
        return 7


from setuptools import setup, find_packages

with open('README.rst') as f:
    README = f.read()

setup(
    name='ddop',
    version='v0.7.5',
    url='https://andreasphilippi.github.io/ddop/',
    license='MIT',
    author='Andreas Philippi',
    author_email='andreas.philippi@uni-wuerzburg.de',
    description='Package for data-driven operations management',
    #long_description=README,
    include_package_data=True,
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=['numpy>=1.18.2', 'scipy>=1.4.1', 'scikit-learn==0.23.0', 'pandas', 'pulp==2.0',
                      'tensorflow>=2.4.1', 'statsmodels>=0.11.1', 'mpmath']
)



from scipy.stats import norm
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from ._base import BaseNewsvendor, ClassicMixin
from ..utils.validation import check_cu_co
from sklearn.utils.validation import check_array
from sklearn.metrics import mean_absolute_error
import numpy as np


class ExponentialSmoothingNewsvendor(BaseNewsvendor, ClassicMixin):
    """
    A SEO newsvendor model with exponential smoothing as underlying forecaster

    Parameters
    ----------
    cu : {array-like of shape (n_outputs,), Number or None}, default=None
       The underage costs per unit. If None, then underage costs are one
       for each target variable
    co : {array-like of shape (n_outputs,), Number or None}, default=None
       The overage costs per unit. If None, then overage costs are one
       for each target variable
    trend : {"add", "mul", "additive", "multiplicative", None}, default=None
        Type of trend component.
    damped_trend : bool, optional (default=None)
        Should the trend component be damped.
    seasonal : {"add", "mul", "additive", "multiplicative", None}, default=None
        Type of seasonal component.
    seasonal_periods : int, default=None
        The number of periods in a complete seasonal cycle, e.g., 4 for
        quarterly data or 7 for daily data with a weekly cycle.
    initialization_method : str, optional
        Method for initialize the recursions. One of:
        * 'estimated'
        * 'heuristic'
        * 'legacy-heuristic'
        * 'known'
        If 'known' initialization is used, then `initial_level`
        must be passed, as well as `initial_trend` and `initial_seasonal` if
        applicable. Default is 'estimated'.
    initial_level : float, optional
        The initial level component. Required if estimation method is "known".
        If set using either "estimated" or "heuristic" this value is used.
        This allows one or more of the initial values to be set while
        deferring to the heuristic for others or estimating the unset
        parameters.
    initial_trend : float, optional
        The initial trend component. Required if estimation method is "known".
        If set using either "estimated" or "heuristic" this value is used.
        This allows one or more of the initial values to be set while
        deferring to the heuristic for others or estimating the unset
        parameters.
    initial_seasonal : array_like, optional
        The initial seasonal component. An array of length `seasonal`
        or length `seasonal - 1` (in which case the last initial value
        is computed to make the average effect zero). Only used if
        initialization is 'known'. Required if estimation method is "known".
        If set using either "estimated" or "heuristic" this value is used.
        This allows one or more of the initial values to be set while
        deferring to the heuristic for others or estimating the unset
        parameters.
    use_boxcox : {True, False, 'log', float}, optional
        Should the Box-Cox transform be applied to the data first? If 'log'
        then apply the log. If float then use lambda equal to float.
    smoothing_level : float, optional
        The alpha value of the simple exponential smoothing, if the value
        is set then this value will be used as the value.
    smoothing_trend : float, optional
        The beta value of the Holt's trend method, if the value is
        set then this value will be used as the value.
    smoothing_seasonal : float, optional
        The gamma value of the holt winters seasonal method, if the value
        is set then this value will be used as the value.
    optimized : bool, optional
        Estimate model parameters by maximizing the log-likelihood
    remove_bias : bool, optional
        Remove bias from forecast values and fitted values by enforcing
        that the average residual is equal to zero.
    method : str, default "L-BFGS-B"
        The minimizer used. Valid options are "L-BFGS-B" (default), "TNC",
        "SLSQP", "Powell", "trust-constr", "basinhopping" (also "bh") and
        "least_squares" (also "ls"). basinhopping tries multiple starting
        values in an attempt to find a global minimizer in non-convex
        problems, and so is slower than the others.

    Attributes
    ----------
    n_outputs_ : int
        The number of outputs when ``fit`` is performed.
    cu_ : ndarray, shape (n_outputs,)
        Validated underage costs.
    co_ : ndarray, shape (n_outputs,)
        Validated overage costs.

    Notes
    -----

    References
    ----------
    .. [1] Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
        and practice. OTexts, 2014.

    Examples
    --------
    >>> from ddop.datasets import load_yaz
    >>> from ddop.newsvendor import ExponentialSmoothingNewsvendor
    >>> from sklearn.model_selection import train_test_split
    >>> X, Y = load_yaz(include_prod=['STEAK'],return_X_y=True)
    >>> Y_train, Y_test = train_test_split(Y, test_size=0.25, shuffle=False, random_state=0)
    >>> mdl = ExponentialSmoothingNewsvendor(cu, co, 'add', False, 'add',7)
    >>> mdl.fit(Y_train)
    >>> mdl.score(Y_test)
    TODO: ADD SCORE
    """

    def __init__(
            self,
            cu=None,
            co=None,
            trend=None,
            damped_trend=False,
            seasonal=None,
            seasonal_periods=None,
            initialization_method="estimated",
            initial_level=None,
            initial_trend=None,
            initial_seasonal=None,
            use_boxcox=False,
            smoothing_level=None,
            smoothing_trend=None,
            smoothing_seasonal=None,
            optimized=True,
            remove_bias=False,
            method="L-BFGS-B"
    ):
        self.cu = cu
        self.co = co
        self.trend = trend
        self.damped_trend = damped_trend
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods
        self.initialization_method = initialization_method
        self.initial_level = initial_level
        self.initial_trend = initial_trend
        self.initial_seasonal = initial_seasonal
        self.use_boxcox = use_boxcox
        self.smoothing_level = smoothing_level
        self.smoothing_trend = smoothing_trend
        self.smoothing_seasonal = smoothing_seasonal
        self.optimized = optimized
        self.remove_bias = remove_bias
        self.method = method
        super().__init__(
            cu=cu,
            co=co)

    def fit(self, y, X=None):
        """Fit the model to the training data y.

        Parameters
        ----------
        y : array-like of shape (n_samples, n_outputs)
            The target values.
        X : array-like of shape (n_samples, n_features), optional (default=None)
            Exogenous variables are ignored

        Returns
        ----------
        self : HoltWintersNewsvendor
            Fitted estimator
        """

        y = check_array(y, ensure_2d=False)

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        self.X_ = X
        self.y_ = y
        self.n_outputs_ = y.shape[1]

        # Check and format under- and overage costs
        self.cu_, self.co_ = check_cu_co(self.cu, self.co, self.n_outputs_)

        forecasters = [ExponentialSmoothing(
            endog=y[:, i],
            trend=self.trend,
            damped_trend=self.damped_trend,
            seasonal=self.seasonal,
            seasonal_periods=self.seasonal_periods,
            initialization_method=self.initialization_method,
            initial_level=self.initial_level,
            initial_trend=self.initial_trend,
            initial_seasonal=self.initial_seasonal,
            use_boxcox=self.use_boxcox) for i in range(self.n_outputs_)]

        fitted_forecasters = [forecasters[i].fit(
            smoothing_level=self.smoothing_level,
            smoothing_trend=self.smoothing_trend,
            smoothing_seasonal=self.smoothing_seasonal,
            optimized=self.optimized,
            remove_bias=self.remove_bias,
            method=self.method) for i in range(self.n_outputs_)]

        error = np.array([y[:, i] - fitted_forecasters[i].fittedvalues for i in range(self.n_outputs_)])
        error_mean = error.mean(axis=1)
        error_std = error.std(axis=1)

        mae = np.array([mean_absolute_error(y[:, i], fitted_forecasters[i].fittedvalues)
                        for i in range(self.n_outputs_)])

        self.safety_buffer_mae = np.array(
            [1.25 * mae[i] * norm.ppf(self.cu_[i] / (self.co_[i] + self.cu_[i])) for i in range(self.n_outputs_)])

        self.safety_buffer_ = np.array(
            [norm(error_mean[i], error_std[i]).ppf(self.cu_[i] / (self.co_[i] + self.cu_[i])) for i in
             range(self.n_outputs_)])

        self.forecasters_ = fitted_forecasters

        return self

    def predict(self, n_steps=1):
        """Predict n time-steps

        Parameters
        ----------
        n_steps : int, default=1
            The number of steps to predict ahead

        Returns
        ----------
        y : array-like of shape (n, n_outputs)
            The predicted values
        """

        forecasts = np.array([forecaster.forecast(n_steps) for forecaster in self.forecasters_]).T
        pred = forecasts + self.safety_buffer_

        return pred



from ..utils.validation import check_cu_co
from sklearn.utils.validation import check_consistent_length
from sklearn.utils.validation import check_array
import numpy as np


def _check_newsvendor_targets(y_true, y_pred):
    """Check that y_true and y_pred belong to the same newsvendor task

    Parameters
    ----------
    y_true : array-like
        The true values
    y_pred : array-like
        The predicted values

    Returns
    -------
    y_true : array-like of shape (n_samples,n_outputs)
        Ground truth (correct) target values.
    y_pred : array-like of shape (n_samples,n_outputs)
        Estimated target values.
    """
    check_consistent_length(y_true, y_pred)
    y_true = check_array(y_true, ensure_2d=False, dtype="numeric")
    y_pred = check_array(y_pred, ensure_2d=False, dtype="numeric")

    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))

    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))

    if y_true.shape[1] != y_pred.shape[1]:
        raise ValueError("y_true and y_pred have different number of output "
                         "({0}!={1})".format(y_true.shape[1], y_pred.shape[1]))

    return y_true, y_pred


def _multiply_cost_weights(x, cu, co):
    if x < 0:
        # underage costs.
        return abs(x) * cu
    else:
        # overage costs
        return x * co


def pairwise_costs(y_true, y_pred, cu, co):
    """ Compute pairwise costs based on the the difference between y_true and y_pred
    and the given underage and overage costs.

    Parameters
    ----------
    y_true : array-like
        The true values
    y_pred : array-like
        The predicted vales
    cu : int or float
        the underage costs per unit.
    co : int or float
        the overage costs per unit.

    Returns
    -------
    costs : ndarray of shape (n_samples, n_outputs)

    Examples
    --------
    >>> from ddop.metrics import pairwise_costs
    >>> y_true = [[2,2], [2,4], [3,6]]
    >>> y_pred = [[1,2], [3,3], [4,7]]
    >>> cu = [2,4]
    >>> co = [1,1]
    >>> pairwise_costs(y_true, y_pred, cu, co)
    array([[2., 0.],
           [1., 4.]
           [1., 1.])
    """
    y_true, y_pred = _check_newsvendor_targets(y_true, y_pred)
    y_diff = y_pred - y_true
    func = np.vectorize(_multiply_cost_weights)
    cu, co = check_cu_co(cu, co, y_true.shape[1])
    costs = func(y_diff, cu, co)
    return costs


def total_costs(y_true, y_pred, cu, co, multioutput="cumulated"):
    """ Compute total costs based on the the difference between y_true and y_pred
    and the given underage and overage costs.

    Parameters
    ----------
    y_true : array-like
        The true values
    y_pred : array-like
        The predicted vales
    cu : int or float
        the underage costs per unit.
    co : int or float
        the overage costs per unit.
    multioutput: {"raw_values", "cumulated"}, default="cumulated"
        Defines aggregating of multiple output values. Default is "cumulated".
         'raw_values' :
            Returns a full set of cost values in case of multioutput input.
         'cumulated' :
            Costs of all outputs are cumulated.
    Returns
    -------
    total_costs :  float or ndarray of floats
        The total costs. If multioutput is ‘raw_values’, then the total costs are returned for each
        output separately. If multioutput is ‘cumulated’, then the cumulated costs of all outputs
        is returned. The total costs are non-negative floating points. The best value is 0.0.

    Examples
    --------
    >>> from ddop.metrics import total_costs
    >>> y_true = [[2,2], [2,4], [3,6]]
    >>> y_pred = [[1,2], [3,3], [4,7]]
    >>> cu = [2,4]
    >>> co = [1,1]
    >>> total_costs(y_true, y_pred, cu, co, multioutput="raw_values")
    array([4,5])
    >>> total_costs(y_true, y_pred, cu, co, multioutput="cumulated")
    9
    """
    costs = pairwise_costs(y_true, y_pred, cu, co)
    total_costs = np.sum(costs, axis=0)

    if multioutput == "raw_values":
        return total_costs

    return np.sum(total_costs)


def average_costs(y_true, y_pred, cu, co, multioutput="uniform_average"):
    """ Compute average costs based on the the difference between y_true and y_pred
    and the given underage and overage costs.

    Parameters
    ----------
    y_true : array-like
        The true values
    y_pred : array-like
        The predicted vales
    cu : int or float
        the underage costs per unit.
    co : int or float
        the overage costs per unit.
    multioutput: {"raw_values", "uniform_average"}, default="raw_values"
        Defines aggregating of multiple output values. Default is "raw_values".
         'raw_values' :
            Returns a full set of cost values in case of multioutput input.
         'uniform_average' :
            Costs of all outputs are averaged with uniform weight.

    Returns
    -------
    costs :  float or ndarray of floats
        The average costs. If multioutput is ‘raw_values’, then the average costs are returned for each
        output separately. If multioutput is ‘uniform_average’, then the average of all output costs is
        returned. The average costs are non-negative floating points. The best value is 0.0.

    Examples
    --------
    >>> from ddop.metrics import average_costs
    >>> y_true = [[2,2], [2,4], [3,6]]
    >>> y_pred = [[1,2], [3,3], [4,7]]
    >>> cu = [2,4]
    >>> co = [1,1]
    >>> average_costs(y_true, y_pred, cu, co, multioutput="raw_values")
    array([1.33.., 1.66..])
    >>> average_costs(y_true, y_pred, cu, co, multioutput="uniform_average")
    1.5
    """
    y_true, y_pred = _check_newsvendor_targets(y_true, y_pred)
    total = total_costs(y_true, y_pred, cu, co, multioutput="raw_values")
    average_costs = total / y_true.shape[0]

    if multioutput == "raw_values":
        return average_costs

    return np.average(average_costs)


def prescriptiveness_score(y_true, y_pred, y_pred_saa, cu, co, multioutput="uniform_average"):
    """ Compute the coefficient of prescriptiveness that is defined as (1 - u/v), where u are the average
    costs between the true and predicted values (y_true,y_pred), and v are the average costs between the
    true values and the predictions obtained by SAA (y_pred_saa, y_pred). The best possible score
    is 1.0 and it can be negative (because the model can be arbitrarily worse).

    Parameters
    ----------
    y_true : array-like
        The true values
    y_pred : array-like
        The predicted vales
    cu : int or float
        the underage costs per unit.
    co : int or float
        the overage costs per unit.
    multioutput: {"raw_values", "uniform_average"}, default="raw_values"
        Defines aggregating of multiple output scores. Default is “uniform_average”.
         'raw_values' :
            Returns a full set of scores in case of multioutput input.
         'uniform_average' :
            Scores of all outputs are averaged with uniform weight.

    Returns
    -------
    score :  float or ndarray of floats
        The prescriptiveness score or ndarray of scores if 'multioutput' is
        'raw_values'.

    Examples
    --------
    >>> from ddop.metrics import prescriptiveness_score
    >>> y_true = [[2,2], [2,4], [3,6]]
    >>> y_pred = [[1,2], [3,3], [4,7]]
    >>> y_pred_saa = [[4,5],[4,5],[4,5]]
    >>> cu = [2,4]
    >>> co = [1,1]
    >>> prescriptiveness_score(y_true, y_pred, cu, co, multioutput="raw_values")
    array([0.2, 0.375])
    >>> prescriptiveness_score(y_true, y_pred, cu, co, multioutput="uniform_average")
    0.2875
    """

    y_true, y_pred = _check_newsvendor_targets(y_true, y_pred)
    y_true, y_pred_saa = _check_newsvendor_targets(y_true, y_pred_saa)

    numerator = average_costs(y_true, y_pred, cu, co)
    denominator = average_costs(y_true, y_pred_saa, cu, co)

    nonzero_denominator = denominator != 0
    nonzero_numerator = numerator != 0
    valid_score = nonzero_denominator & nonzero_numerator
    output_scores = np.ones([y_true.shape[1]])
    output_scores[valid_score] = 1 - (numerator[valid_score] /
                                      denominator[valid_score])
    # arbitrary set to zero to avoid -inf scores, having a constant
    # y_true is not interesting for scoring anyway
    # output_scores[nonzero_numerator & ~nonzero_denominator] = 0.

    if multioutput == "raw_values":
        return output_scores

    return np.average(output_scores)
